/*******************************************************************************
 * Copyright (c) 2000, 2008 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *     
 * Port to the D programming language:
 *     Jacob Carlborg <doob@me.com>
 *******************************************************************************/
module dwt.dnd.TreeDropTargetEffect;

import dwt.dwthelper.utils;

import dwt.DWT;
import dwt.graphics.Point;
import dwt.graphics.Rectangle;
import dwt.widgets.Event;
import dwt.widgets.Tree;
import dwt.widgets.TreeItem;

import dwt.dnd.DND;
import dwt.dnd.DropTarget;
import dwt.dnd.DropTargetEffect;
import dwt.dnd.DropTargetEvent;

/**
 * This class provides a default drag under effect (eg. select, insert, scroll and expand) 
 * when a drag occurs over a <code>Tree</code>.
 * 
 * <p>Classes that wish to provide their own drag under effect for a <code>Tree</code>
 * can extend the <code>TreeDropTargetEffect</code> class and override any applicable methods 
 * in <code>TreeDropTargetEffect</code> to display their own drag under effect.</p>
 *
 * Subclasses that override any methods of this class must call the corresponding
 * <code>super</code> method to get the default drag under effect implementation.
 *
 * <p>The feedback value is either one of the FEEDBACK constants defined in 
 * class <code>DND</code> which is applicable to instances of this class, 
 * or it must be built by <em>bitwise OR</em>'ing together 
 * (that is, using the <code>int</code> "|" operator) two or more
 * of those <code>DND</code> effect constants. 
 * </p>
 * <p>
 * <dl>
 * <dt><b>Feedback:</b></dt>
 * <dd>FEEDBACK_SELECT, FEEDBACK_INSERT_BEFORE, FEEDBACK_INSERT_AFTER, FEEDBACK_EXPAND, FEEDBACK_SCROLL</dd>
 * </dl>
 * </p><p>
 * Note: Only one of the styles FEEDBACK_SELECT, FEEDBACK_INSERT_BEFORE or
 * FEEDBACK_INSERT_AFTER may be specified.
 * </p>
 * 
 * @see DropTargetAdapter
 * @see DropTargetEvent
 * @see <a href="http://www.eclipse.org/swt/">Sample code and further information</a>
 * 
 * @since 3.3
 */
public class TreeDropTargetEffect : DropTargetEffect {
    static const int SCROLL_HYSTERESIS = 150; // milli seconds
    static const int EXPAND_HYSTERESIS = 1000; // milli seconds

    int currentEffect = DND.FEEDBACK_NONE;
    TreeItem currentItem;

    TreeItem insertItem = null;
    bool insertBefore = false;

    TreeItem scrollItem;
    long scrollBeginTime;

    TreeItem expandItem;
    long expandBeginTime;
    
//  DataBrowserCallbacks callbacks = null;
    
    int acceptDragProc(int theControl, int itemID, int property, int theRect, int theDrag) {
        return (currentEffect & DND.FEEDBACK_SELECT) !is 0 ? 1 : 0;
    }

//  static Callback AcceptDragProc;
//  static {
//      AcceptDragProc = new Callback(TreeDropTargetEffect.class, "AcceptDragProc", 5); //$NON-NLS-1$
//      int acceptDragProc = AcceptDragProc.getAddress();
//      if (acceptDragProc is 0) DWT.error(DWT.ERROR_NO_MORE_CALLBACKS);
//  }

//  static int AcceptDragProc(int theControl, int itemID, int property, int theRect, int theDrag) {
//      DropTarget target = FindDropTarget(theControl, theDrag);
//      if (target is null) return 0;
//      return (target.feedback & DND.FEEDBACK_SELECT) !is 0 ? 1 : 0;
//  }
    
//  static DropTarget FindDropTarget(int theControl, int theDrag) {
//      if (theControl is 0) return null;
//      Display display = Display.findDisplay(Thread.currentThread());
//      if (display is null || display.isDisposed()) return null;
//      Widget widget = display.findWidget(theControl);
//      if (widget is null) return null;
//      return (DropTarget)widget.getData(DND.DROP_TARGET_KEY); 
//  }

    /**
     * Creates a new <code>TreeDropTargetEffect</code> to handle the drag under effect on the specified 
     * <code>Tree</code>.
     * 
     * @param tree the <code>Tree</code> over which the user positions the cursor to drop the data
     */
    public this(Tree tree) {
        super(tree);
    }

    int checkEffect(int effect) {
        // Some effects are mutually exclusive.  Make sure that only one of the mutually exclusive effects has been specified.
        if ((effect & DND.FEEDBACK_SELECT) !is 0) effect = effect & ~DND.FEEDBACK_INSERT_AFTER & ~DND.FEEDBACK_INSERT_BEFORE;
        if ((effect & DND.FEEDBACK_INSERT_BEFORE) !is 0) effect = effect & ~DND.FEEDBACK_INSERT_AFTER;
        return effect;
    }

    /**
     * This implementation of <code>dragEnter</code> provides a default drag under effect
     * for the feedback specified in <code>event.feedback</code>.
     * 
     * For additional information see <code>DropTargetAdapter.dragEnter</code>.
     * 
     * Subclasses that override this method should call <code>super.dragEnter(event)</code>
     * to get the default drag under effect implementation.
     *
     * @param event  the information associated with the drag enter event
     * 
     * @see DropTargetAdapter
     * @see DropTargetEvent
     */
    public void dragEnter(DropTargetEvent event) {
//      if (callbacks is null) {
//          Tree table = (Tree) control;
//          DataBrowserCallbacks callbacks = new DataBrowserCallbacks ();
//          OS.GetDataBrowserCallbacks (table.handle, callbacks);
//          callbacks.v1_acceptDragCallback = AcceptDragProc.getAddress();
//          OS.SetDataBrowserCallbacks(table.handle, callbacks);
//      }
        insertItem = null;
        expandBeginTime = 0;
        expandItem = null;
        scrollBeginTime = 0;
        scrollItem = null;
    }

    /**
     * This implementation of <code>dragLeave</code> provides a default drag under effect
     * for the feedback specified in <code>event.feedback</code>.
     * 
     * For additional information see <code>DropTargetAdapter.dragLeave</code>.
     * 
     * Subclasses that override this method should call <code>super.dragLeave(event)</code>
     * to get the default drag under effect implementation.
     *
     * @param event  the information associated with the drag leave event
     * 
     * @see DropTargetAdapter
     * @see DropTargetEvent
     */
    public void dragLeave(DropTargetEvent event) {
        Tree tree = cast(Tree) control;
        if (insertItem !is null) {
            setInsertMark(tree, null, false);
            insertItem = null;
        }
        expandBeginTime = 0;
        expandItem = null;
        scrollBeginTime = 0;
        scrollItem = null;
    }

    /**
     * This implementation of <code>dragOver</code> provides a default drag under effect
     * for the feedback specified in <code>event.feedback</code>.
     * 
     * For additional information see <code>DropTargetAdapter.dragOver</code>.
     * 
     * Subclasses that override this method should call <code>super.dragOver(event)</code>
     * to get the default drag under effect implementation.
     *
     * @param event  the information associated with the drag over event
     * 
     * @see DropTargetAdapter
     * @see DropTargetEvent
     * @see DND#FEEDBACK_SELECT
     * @see DND#FEEDBACK_INSERT_BEFORE
     * @see DND#FEEDBACK_INSERT_AFTER
     * @see DND#FEEDBACK_SCROLL
     */
    public void dragOver(DropTargetEvent event) {
        Tree tree = cast(Tree) control;
        TreeItem item = cast(TreeItem)getItem(tree, event.x, event.y);
        int effect = checkEffect(event.feedback);       
        if ((effect & DND.FEEDBACK_EXPAND) is 0) {
            expandBeginTime = 0;
            expandItem = null;
        } else {
            if (item !is null && item.opEquals(expandItem) && expandBeginTime !is 0) {
                if (System.currentTimeMillis() >= expandBeginTime) {
                    if (item.getItemCount() > 0 && !item.getExpanded()) {
                        Event e = new Event();
                        e.x = event.x;
                        e.y = event.y;
                        e.item = item;
                        e.time = cast(int) System.currentTimeMillis();
                        tree.notifyListeners(DWT.Expand, e);
                        if (item.isDisposed()) return;
                        item.setExpanded(true);
                    }
                    expandBeginTime = 0;
                    expandItem = null;
                }
            } else {
                expandBeginTime = System.currentTimeMillis() + EXPAND_HYSTERESIS;
                expandItem = item;
            }
        }
        
        if ((effect & DND.FEEDBACK_SCROLL) is 0) {
            scrollBeginTime = 0;
            scrollItem = null;
        } else {
            if (item !is null && item.opEquals(scrollItem)  && scrollBeginTime !is 0) {
                if (System.currentTimeMillis() >= scrollBeginTime) {
                    Rectangle area = tree.getClientArea();
                    int headerHeight = tree.getHeaderHeight();
                    int itemHeight= tree.getItemHeight();
                    Point pt = new Point(event.x, event.y);
                    pt = tree.getDisplay().map(null, tree, pt);
                    TreeItem nextItem_ = null;
                    if (pt.y < area.y + headerHeight + 2 * itemHeight) {
                        nextItem_ = previousItem(tree, item);
                    }
                    if (pt.y > area.y + area.height - 2 * itemHeight) {
                        nextItem_ = nextItem(tree, item);
                    }
                    if (nextItem_ !is null) tree.showItem(nextItem_);
                    scrollBeginTime = 0;
                    scrollItem = null;
                }
            } else {
                scrollBeginTime = System.currentTimeMillis() + SCROLL_HYSTERESIS;
                scrollItem = item;
            }
        }
        
        if ((effect & DND.FEEDBACK_INSERT_AFTER) !is 0 ||
            (effect & DND.FEEDBACK_INSERT_BEFORE) !is 0) {
            if (currentItem !is item || 
                 ((effect & DND.FEEDBACK_INSERT_AFTER) !is (currentEffect & DND.FEEDBACK_INSERT_AFTER)) ||
                 ((effect & DND.FEEDBACK_INSERT_BEFORE) !is (currentEffect & DND.FEEDBACK_INSERT_BEFORE))) { 
                setInsertMark(tree, item, (effect & DND.FEEDBACK_INSERT_BEFORE) !is 0);
                currentEffect = effect;
                currentItem = item;
            }
        } else {
            setInsertMark(tree, null, false);
        }
        // save current effect for selection feedback
        (cast(DropTarget)event.widget).feedback = effect;
    }

    void setInsertMark(Tree tree, TreeItem item, bool before) {
        if (item is insertItem && before is insertBefore) return;
        insertItem = item;
        insertBefore = before;
        tree.setInsertMark(item, before);
    }
}
